/********************************** cgm.c ***********************************/

#include <stdio.h>
#include <alloc.h>
#include <stdarg.h>
#include <string.h>
#include "cgmasm.h"
#include "cgm.h"



FILE *_cgmout;      /* output file pointer */
int _color_mode_out = 0;
unsigned int _maxpolybuf    = 4004;
unsigned int _outfilebuffer = 20000;

/* static variables for the poly element functions */
static int index;
static int flag = 0;
static int *polybuf;



/************************ begin_metafile defines ******************************/


#define INTEGER     0
#define REAL        1 

/* Valid INTEGER PRECISION, INDEX PRECISION,COLOUR PRECISION, COLOUR INDEX
   PRECISION or MAXIMUM COLOUR INDEX values are 8,16,24 and 32 bits       */

/* REAL PRECISION parameters: */

/* Parameter 1 */
#define FLOAT 0
#define FIXEDPT 1

/* Legal combinations of values for parameters 2 and 3 */

/*         P1         P2      P3            RESULT             */
/*       FLOAT         9      23     32-bit floating point     */
/*       FLOAT        12      53     64-bit floating point     */
/*       FIXEDPT      16      16     32-bit fixed point        */
/*       FIXEDPT      32      32     64-bit fixed point        */

/* SCALING MODE parameters */

/* Parmeter 1 */
#define ABSTRACT 0
#define METERIC  1

/* Parmeter 2 is (real) meteric scaling factor, ignored of P1 = ABSTRACT */

/* COLOUR SELECTION MODE parameter */
#define INDEXED 0
#define DIRECT  1

/* Valid parameter values for LINE WIDTH, MARKER SIZE and EDGE WIDTH MODES
   are :  */

#define ABSOLUTE  0
#define SCALED    1

/* Valid VDC INTEGER PRECISION values are 16,24 and 32 bits (not 8 bits)*/



/****************************************************************************/


void cgm_element(int op,...)
{
   int n,k,buf[14];
   va_list ap;

   n = (op & 0x001F)/2;
   buf[0] = op;
   va_start(ap,op);
   for ( k = 1; k <= n; k++)
   {
      buf[k] = va_arg(ap,int);
   }
   va_end(ap);
   n++;
   cgm_swab( n,buf);
   fwrite(buf,2,n,_cgmout);
}


void cgm_color(int type,...)
{
   int n,k,buf[14];
   va_list ap;
   buf[0] = type;
   if(_color_mode_out)
   {
      buf[0] += 6;
      n = 3;
   }
   else
   {
      buf[0] += 2;
      n = 1;
   }


   va_start(ap,type);
   for ( k = 1; k <= n; k++)
   {
      buf[k] = va_arg(ap,int);
   }
   va_end(ap);
   n++;
   cgm_swab( n,buf);
   fwrite(buf,2,n,_cgmout);
}



int cgm_str(int op,...)
{
    unsigned int count,temp;
    int buf[8];
    int n,k;
    char *cpt;
    va_list ap;

   n = (op & 0x001F)/2 +1;
   va_start(ap,op);
   for ( k = 2; k <= n; k++)
   {
      buf[k] = va_arg(ap,int);
   }
   cpt = va_arg(ap,char *);
   va_end(ap);

   count = strlen(cpt);
   if ( count > 254)
       return(0);
   temp = n*2 + strlen(cpt)  -1;
   if(temp > 30)
   {
      buf[0] = op | 31;
      buf[1] = temp;
      cgm_swab(n+1,buf);
      fwrite(buf,2,n +1,_cgmout);
   }
   else
   {
      buf[1] = (0xffe0 & op) | temp;
      cgm_swab(n,&buf[1]);
      fwrite(&buf[1],2,n,_cgmout);
   }
    fwrite(&count,1,1,_cgmout);
    if(temp & 0x0001)
       count++;
    fwrite(cpt,1,count,_cgmout); 
    return(1);
}


int cgm_begin_metafile(char *filename,char *begindescr,char *metadescr,
                       int maxcolorindex,int *mincolor,int *maxcolor,
                       char *font_list[],CHARSET *charlist,int charcoding)

{

/* begin_metafile returns zero on success and a error number on failure    */
/*     return error 1 --- begindescr greater than 254 characters.          */
/*     return error 2 --- file could not be opened for writting.           */
/*     return error 3 --- metadescr greater than 254 characters.           */
/*     return error 4 --- Not enough memory for file buffer               */
/*     return error 5 --- font list Memory Error                           */
/*     return error 6 --- font list > MAXCBUF charcters                    */

    unsigned int temp,j,k;
    int buf[256];
    unsigned int count;
    char *pchr;
    int *pint;
    flag = 0;

    _cgmout = fopen(filename,"wb");
    if (_cgmout == NULL) return(2);

    if ( setvbuf(_cgmout,NULL,_IOFBF,_outfilebuffer) != 0)
    {
        fclose(_cgmout);
        return(4);
    }

    if(!cgm_str(BEGMF,begindescr))
    {
       fclose(_cgmout);
       return(1);
    }

    cgm_element(MFVERSION +2,1);  /* version 1 */

    if(!cgm_str(MFDESC,metadescr))
    {
       fclose(_cgmout);
       return(3);
    }

    cgm_element(VDCTYPE +2,INTEGER);
    cgm_element(INTEGERPREC +2,16); 
    cgm_element(REALPREC +6,FIXEDPT,16,16);

    cgm_element(INDEXPREC +2,16);
    cgm_element(COLRPREC +2,16); 

    if( _color_mode_out == 0)
    {
       cgm_element(COLRINDEXPREC +2,16);
       cgm_element(MAXCOLRINDEX +2,maxcolorindex);
    }
    
    cgm_element(COLRVALUEEXT +12,mincolor[0],mincolor[1],mincolor[2],
                                 maxcolor[0],maxcolor[1],maxcolor[2]);

    /* metafile element list */

       buf[0] = MFELEMLIST + 6;
       buf[1] = 1;                        /* number of elements */
       buf[2] = -1;
       buf[3] = 1;                     /* drawing set + control */
       cgm_swab( 4,buf);
       fwrite(buf,2,4,_cgmout); 

    /* font list */
    if ( (font_list[0] != NULL) && (font_list[0][0] != '\0'))
    {  /* write font element */
       pchr = NULL;
       pchr = (char *) malloc(MAXFONTBUF);
       if( pchr == NULL)
       {
          fclose(_cgmout);        
          return(5);
       }

       temp = 0;
       for ( j = 0;;j++)
       {
          if(font_list[j] == NULL)  break;
          count = strlen(font_list[j]);
          if (count <= 0) break;
          if ( (count + temp + 1) > (MAXFONTBUF -1))
          {
              free(pchr);
              fclose(_cgmout);
              return(5);
          }
          pchr[temp++] = count;
          for (k = 0;k < count;k++)
          {
             pchr[temp++] = font_list[j][k];
          }

       }

       buf[0] = FONTLIST + 31;
       buf[1] = temp;                     /* number bytes */
       cgm_swab(2,buf);
       fwrite(buf,2,2,_cgmout);            /* write font list header */

       pchr[temp] = 0;
       if (temp & 0x0001) temp++;
       fwrite(pchr,1,temp,_cgmout);   /* write font names */
       free(pchr);
    }

    /* character set list */
     if( charlist != NULL)
     {
        k =0;
       count = 0;
       buf[0] = CHARSETLIST + 31;
       pint =  &buf[2];
       while(charlist[k].tail != NULL)
       {
          temp = strlen(charlist[k].tail);
          if ( (count + temp) > 500) break;
          if (temp > 254) break;
          *pint = charlist[k].type;
          cgm_swab(1,pint);
          pint++;
          pchr  = (char *) pint;
          *pchr = temp;
          pchr++;
          for (j = 0;j < temp;j++)
          {
             *pchr = charlist[k].tail[j];
             pchr++;
          }
          *pchr = 0;
          pint = (int *) pchr;
          count += 3 + temp;
          k++;
       }
       buf[1] = count;                     /* number bytes */
       cgm_swab(2,buf);
       if(count &0x0001)
          count++;
       fwrite(buf,1,count + 4,_cgmout);   /* write charsetlist */
       cgm_element(CHARCODING +2,charcoding);
    }
    return(0);
}

int cgm_begin_picture(char *descr,FIXED meteric,int X1,int Y1,int X2,int Y2,
                      int *bgcolor)
{
    int  k, *ptr;
    if (!cgm_str(BEGPIC,descr))
    {
        fclose(_cgmout);
        return (5);
    }

    /* SCALEMODE --- scale factor not used in abstract mode  */

    if (meteric <= 0)
         k = ABSTRACT;
    else
         k = METERIC;
    ptr = (int *) &meteric;
    cgm_element(SCALEMODE +6,k,ptr[1],ptr[0]);

    if( _color_mode_out == 0)
       cgm_element(COLRMODE +2,INDEXED);
    else
       cgm_element(COLRMODE +2,DIRECT);

    cgm_element(LINEWIDTHMODE +2,ABSOLUTE);
    cgm_element(MARKERSIZEMODE +2,ABSOLUTE);
    cgm_element(EDGEWIDTHMODE +2,ABSOLUTE);

    cgm_element(VDCEXT +8,X1,Y1,X2,Y2);


    /* background color */
    if (bgcolor != NULL)
    {  /* write backgound color element */
       cgm_element(BACKCOLR +6,bgcolor[0],bgcolor[1],bgcolor[2]);
    }
    cgm_element(BEGPICBODY);
    cgm_element(VDCINTEGERPREC +2,16);        /* 16 bits  */
    return(0);
}



/*   cgm_end_metafile()  returns zero on success and a one on failure   */

int cgm_end_metafile(void)
{
    cgm_element(ENDMF);
    if ( fclose(_cgmout) != 0)
        return(0);
    return(1);
}

int cgm_polybuf(int op, int num, int *far polybuffer)
{
   int buf[4], k;
   if ( num > 16000)
      return(0);
   cgm_swab(2*num,polybuffer); /* swap poly points */
   buf[0] = op + 31;
   if( num <= 8000)
   {
      buf[1] = 4*num;
      cgm_swab(2,buf);
      fwrite(buf,2,2,_cgmout);        /* write header + partition */
      fwrite(polybuffer,4,num,_cgmout);        /* write points */
      return(num);
   }
   else
   {
      buf[1] = (4*8000) | 0x8000;
      cgm_swab(2,buf);
      fwrite(buf,2,2,_cgmout);              /* write header + partition */
      fwrite(polybuffer,4,8000,_cgmout);        /* write points */
   }
   k = num - 8000;

   buf[0] = 4*k;   
   cgm_swab(1,buf);
   fwrite(buf,2,1,_cgmout);                     /* write  partition */
   fwrite(&polybuffer[2*8000],4,k,_cgmout);        /* write points */
   return(num);
}

int cgm_begin_poly(int op,int x,int y)
{
    polybuf = NULL;
    polybuf = malloc(2*_maxpolybuf);
    if (polybuf == NULL)
       return(0);
    polybuf[0] = op + 31;
    polybuf[2] = x;
    polybuf[3] = y;
    index = 4;
    return(1);
}

void cgm_poly(int x,int y)
{
    if ( (index +1) < _maxpolybuf)
    {
        polybuf[index++] = x;
        polybuf[index++] = y;
    }
    else
    {
        if (flag == 0) 
        {
            flag = 1;
            polybuf[1] = 0x8000 | (index << 1) - 4;
        }
        else
        {
            polybuf[0] = 0x8000 | (index << 1) - 2;
        }

        cgm_swab(index,polybuf);
        fwrite(polybuf,2,index,_cgmout);        /* write poly partition */
        polybuf[1] = x;
        polybuf[2] = y;
        index = 3;
    }
}
 

void cgm_end_poly(void)
{
    if (flag == 0)
        polybuf[1] = (index << 1) - 4;
    else
        polybuf[0] = (index << 1) - 2;
    cgm_swab(index,polybuf);
    fwrite(polybuf,2,index,_cgmout);        /* write poly partition */
    flag = 0;
    free(polybuf);
    polybuf = NULL;
}


/* Continuous alignment values are real numbers */

void cgm_text_alignment(int hai,int vai,FIXED chor,FIXED cvert)
{
    int *ptr, buf[7];
    buf[0] = TEXTALIGN + 12;
    buf[1] = hai;
    buf[2] = vai;
    ptr = (int *) &chor;
    buf[3] = ptr[1];
    buf[4] = ptr[0];
    ptr = (int *) &cvert;
    buf[5] = ptr[1];
    buf[6] = ptr[0];
    cgm_swab(7,buf);
    fwrite(buf,2,7,_cgmout);        /* write text align element */
}


int cgm_color_table( int sindex,int num,int *table)
{
    int temp,*buf,j,k;
    buf = NULL;
    buf = (int *) malloc(6*num + 6);
    if (buf == NULL)
       return(0);
    temp = 6*num + 2;
    buf[0] = COLRTABLE + 31;
    buf[1] = temp;
    buf[2] = sindex;
    for ( j = 2,k = 0;k < num;k++)
    {
        buf[++j] = table[3*k];
        buf[++j] = table[3*k+1];
        buf[++j] = table[3*k+2];
    }
    j++;
    cgm_swab(j,buf);
    fwrite(buf,2,j,_cgmout);        /* write color table element */
    free(buf);
    return(1);
}





