/******************************* cgm.h **************************************/


#define MAXFONTBUF 512

typedef long FIXED;

struct _charset
{
   int type;
   char *tail;
};

typedef struct _charset CHARSET;

extern FILE *_cgmout;
extern int _color_mode_out;
extern unsigned int _maxpolybuf;
extern unsigned int _outfilebuffer;

/*  List of Binary Encoding Metafile Element Codes .                */
/*  OPCODES are Clear Text Derived Names                            */

#define NOOP               0x0000
#define BEGMF              0x0020
#define ENDMF              0x0040
#define BEGPIC             0x0060
#define BEGPICBODY         0x0080
#define ENDPIC             0x00A0

#define MFVERSION          0x1020
#define MFDESC             0x1040
#define VDCTYPE            0x1060
#define INTEGERPREC        0x1080
#define REALPREC           0x10A0
#define INDEXPREC          0x10C0
#define COLRPREC           0x10E0
#define COLRINDEXPREC      0x1100
#define MAXCOLRINDEX       0x1120
#define COLRVALUEEXT       0x1140
#define MFELEMLIST         0x1160
#define FONTLIST           0x11A0
#define CHARSETLIST        0x11C0
#define CHARCODING         0x11E0



#define SCALEMODE          0x2020
#define COLRMODE           0x2040
#define LINEWIDTHMODE      0x2060
#define MARKERSIZEMODE     0x2080
#define EDGEWIDTHMODE      0x20A0 
#define VDCEXT             0x20C0
#define BACKCOLR           0x20E0

#define VDCINTEGERPREC     0x3020
#define AUXCOLR            0x3060
#define TRANSPARENCY       0x3080
#define CLIPRECT           0x30A0
#define CLIP               0x30C0

#define LINE               0x4020
#define DISJLINE           0x4040
#define MARKER             0x4060
#define TEXT               0x4080
#define RESTRTEXT          0x40A0
#define APNDTEXT           0x40C0
#define POLYGON            0x40E0
#define POLYGONSET         0x4100
#define CELLARRAY          0x4120
#define GDP                0x4140
#define RECT               0x4160
#define CIRCLE             0x4180
#define ARC3PT             0x41A0
#define ARC3PTCLOSE        0x41C0
#define ARCCTR             0x41E0
#define ARCCTRCLOSE        0x4200
#define ELLIPSE            0x4220
#define ELLIPARC           0x4240
#define ELLIPARCCLOSE      0x4260

#define LINETYPE           0x5040
#define LINEWIDTH          0x5060
#define LINECOLR           0x5080
#define MARKERTYPE         0x50C0
#define MARKERSIZE         0x50E0
#define MARKERCOLR         0x5100
#define TEXTFONTINDEX      0x5140
#define TEXTPREC           0x5160
#define CHAREXPAN          0x5180
#define CHARSPACE          0x51A0
#define TEXTCOLR           0x51C0
#define CHARHEIGHT         0x51E0
#define CHARORI            0x5200
#define TEXTPATH           0x5220
#define TEXTALIGN          0x5240
#define CHARSETINDEX       0x5260
#define ALTCHARSETINDEX    0x5280
#define INTSTYLE           0x52C0
#define FILLCOLR           0x52E0
#define HATCHINDEX         0x5300
#define PATINDEX           0x5320
#define EDGETYPE           0x5360
#define EDGEWIDTH          0x5380
#define EDGECOLR           0x53A0
#define EDGEVIS            0x53C0
#define FILLREFPT          0x53E0
#define PATTABLE           0x5400
#define PATSIZE            0x5420
#define COLRTABLE          0x5440
#define ESCAPE             0x6020 
#define MESSAGE            0x7020 
#define APPLDATA           0x7040 



/* CGM functions with one parameter */

unsigned int cgm_strlen(char far *string);
void cgm_element(int op,...);
void cgm_color(int type,...);

/* Default Color Index List */

#define  BACKGROUND      0  /* device dependent */
#define  FOREGROUND      1  /* device dependent */


#define cgm_transparency(state)       cgm_element(TRANSPARENCY +2,(int) state)
#define cgm_clip(state)               cgm_element(CLIP +2,(int) state)
#define cgm_line_type(type)           cgm_element(LINETYPE +2,(int) type)
#define cgm_line_width(width)         cgm_element(LINEWIDTH +2,(int) width)
#define cgm_marker_type(type)         cgm_element(MARKERTYPE +2,(int) type)
#define cgm_marker_size(size)         cgm_element(MARKERSIZE +2,(int) size)
#define cgm_interior_style(style)     cgm_element(INTSTYLE +2,(int) style)
#define cgm_hatch_index(index)        cgm_element(HATCHINDEX +2,(int) index)
#define cgm_edge_type(type)           cgm_element(EDGETYPE +2,(int) type)
#define cgm_edge_width(width)         cgm_element(EDGEWIDTH +2,(int) width)
#define cgm_edge_vis(mode)            cgm_element(EDGEVIS +2,(int) mode)
#define cgm_pattern_index(index)      cgm_element(PATINDEX +2,(int) index)
#define cgm_text_font_index(index)    cgm_element(TEXTFONTINDEX +2,(int) index)
#define cgm_char_height(height)       cgm_element(CHARHEIGHT +2,(int) height)
#define cgm_text_precision(state)     cgm_element(TEXTPREC +2,(int) state)
#define cgm_text_path(state)          cgm_element(TEXTPATH +2,(int) state)
#define cgm_char_set_index(index)     cgm_element(CHARSETINDEX +2,(int) index)
#define cgm_alt_char_set_index(index) cgm_element(ALTCHARSETINDEX+2,(int)index)




int cgm_color_table( int sindex,int num,int *table);

/*  polyline, polymarker and polygon functions */

int cgm_polybuf(int op, int num, int *far polybuf);

int cgm_begin_poly(int op,int x,int y);
void cgm_poly(int x,int y);
void cgm_end_poly(void);
                                     

#define cgm_bpolyline(num,buffer)     cgm_polybuf(LINE,num,buffer)
#define cgm_bpolymarker(num,buffer)   cgm_polybuf(MARKER,num,buffer)
#define cgm_bpolygon(num,buffer)      cgm_polybuf(POLYGON,num,buffer)
#define cgm_bdisjline(num,buffer)     cgm_polybuf(DISJLINE,num,buffer)

#define cgm_begin_polyline(x,y)     cgm_begin_poly(LINE,x,y)
#define cgm_polyline(x,y)           cgm_poly(x,y)
#define cgm_end_polyline()           cgm_end_poly()
#define cgm_begin_polymarker(x,y)   cgm_begin_poly(MARKER,x,y)
#define cgm_polymarker(x,y)         cgm_poly(x,y)
#define cgm_end_polymarker()         cgm_end_poly()
#define cgm_begin_polygon(x,y)      cgm_begin_poly(POLYGON,x,y)
#define cgm_polygon(x,y)            cgm_poly(x,y)
#define cgm_end_polygon()            cgm_end_poly()
#define cgm_begin_disjline(x,y)     cgm_begin_poly(DISJLINE,x,y)
#define cgm_disjline(x,y)           cgm_poly(x,y)
#define cgm_end_disjline()         cgm_end_poly()



int cgm_begin_metafile(char *filename,char *begindescr,char *metadescr,
                       int maxcolorindex,int *mincolor,int *maxcolor,
                       char *font_list[],CHARSET *charlist,int charcoding);
int cgm_begin_picture(char *descr,FIXED meteric,int X1,int Y1 ,int X2,int Y2,
                       int *bgcolor);

#define cgm_end_picture() cgm_element(ENDPIC)

int cgm_end_metafile(void);


#define cgm_line(x1,y1,x2,y2)   \
  cgm_element(LINE +8,(int) x1,(int) y1,(int) x2,(int) y2)
#define cgm_rectangle(x1,y1,x2,y2)  \
  cgm_element(RECT +8,(int) x1,(int) y1,(int) x2,(int) y2)
#define cgm_clip_rectangle(x1,y1,x2,y2) \
  cgm_element(CLIPRECT +8,(int) x1,(int) y1,(int) x2,(int) y2)
#define cgm_char_orientation(ux,uy,bx,by) \
  cgm_element(CHARORI +8,(int) ux,(int) uy,(int) bx,(int) by)

#define cgm_marker(x,y)   cgm_element(MARKER + 4,(int) x,(int) y)


#define cgm_circle(cx,cy,r) \
  cgm_element(CIRCLE +6,(int) cx,(int) cy,(int) r)
#define cgm_circular_arc_3pt(sx,sy,mx,my,ex,ey) \
  cgm_element(ARC3PT +12,(int) sx,(int) sy,(int) mx,(int) my,(int) ex,(int) ey)
#define cgm_circular_arc_3pt_close(sx,sy,mx,my,ex,ey,flag) \
  cgm_element(ARC3PTCLOSE +14,(int) sx,(int) sy,(int) mx,(int) my,(int) ex,\
              (int) ey,(int) flag)
#define cgm_ellipse(cx,cy,x1,y1,x2,y2) \
  cgm_element(ELLIPSE +12,(int) cx,(int) cy,(int) x1,(int) y1,(int) x2,(int) y2)
#define cgm_circular_arc_center(cx,cy,sx,sy,ex,ey,r) \
  cgm_element(ARCCTR +14,(int) cx,(int) cy,(int) sx,(int) sy,(int) ex, \
              (int) ey,(int) r)
#define cgm_circular_arc_center_close(cx,cy,sx,sy,ex,ey,r,type) \
  cgm_element(ARCCTRCLOSE +16,(int) cx,(int) cy,(int) sx,(int) sy,(int) ex,\
              (int) ey,(int) r,(int) type)
#define cgm_elliptical_arc(cx,cy,x1,y1,x2,y2,sx,sy,ex,ey) \
  cgm_element(ELLIPARC +20,(int) cx,(int) cy,(int) x1,(int) y1,(int) x2,\
              (int) y2,(int) sx,(int) sy,(int) ex,(int) ey)
#define cgm_elliptical_arc_close(cx,cy,x1,y1,x2,y2,sx,sy,ex,ey,type) \
  cgm_element(ELLIPARCCLOSE +22,(int) cx,(int) cy,(int) x1,(int) y1,(int) x2,\
              (int) y2,(int) sx,(int) sy,(int) ex,(int) ey,(int) type)


void cgm_text_alignment(int hai,int vai,FIXED chor, FIXED cvert);


int cgm_str(int op,...);

#define cgm_text(x,y,flag,str) \
  cgm_str(TEXT + 6,(int) x,(int) y,(int) flag,(char *) str)
#define cgm_restricted_text(w,h,x,y,flag,str) \
  cgm_str(RESTRTEXT + 10,(int) w,(int) h,(int) x,(int) y, \
           (int) flag,(char *) str)
#define cgm_append_text(p1,str) cgm_str(APNDTEXT +2, (int) p1,(char *) str)
#define cgm_message(p1,str) cgm_str(MESSAGE + 2,(int) p1,(char *) str) 

/******************** Extended Functions ************************************/
void cgm_gdp(int id,int num,int *plist,int numchr,char *data); 
void cgm_polygon_set(int num, int *array);
int cgmp1str(int op,int p1,int count,char *string);
#define cgm_escape(p1,num,string) cgmp1str(ESCAPE,p1,num,string) 
#define cgm_application_data(p1,num,string) cgmp1str(APPLDATA,p1,num,string) 
void cgm1real(int op,FIXED p);
#define cgm_char_expansion(factor)    cgm1real(CHAREXPAN,factor)
#define cgm_char_spacing(factor)      cgm1real(CHARSPACE,factor)
/****************************************************************************/



/* Text precision */

#define STRING        0
#define CHARACTER     1
#define STROKE        2

/* Text path */

#define TP_RIGHT      0
#define TP_LEFT       1
#define TP_UP         2
#define TP_DOWN       3




/* Line and edge type */


#define SOLID        1
#define DASH         2  
#define DOT          3
#define DASHDOT      4
#define DASHDOTDOT   5



/* Text Final/non-final Flag */

#define NON_FINAL        0 
#define FINAL           1 


/*  Text Horizontal Alignment */

#define NORMAL          0
#define LEFT            1
#define CENTER          2
#define RIGHT           3

/*  Text Vertical Alignment */

#define NORMAL          0
#define TOP             1
#define CAP             2
#define HALF            3
#define BASE            4
#define BOTTOM          5


/* Marker Type */

#define MDOT      1
#define PLUS      2
#define ASTERISK  3
#define MCIRCLE   4
#define CROSS     5


/* interior style */

#define HOLLOW         0
#define SOLID          1
#define PATTERN        2
#define HATCH          3
#define EMPTY          4


/* hatch index    */ 

#define  HOR      1
#define  VERT     2
#define  PSLOPE   3
#define  NSLOPE   4
#define  VHSLANT  5
#define  LRSLANT  6


#define PIE            0
#define CHORD          1

#define OFF            0
#define ON             1

/* message flags */
#define NO_ACTION      0
#define ACTION         1   

/* polygon set flags */

#define INVISIBLE 0
#define VISIBLE 1
#define CLOSE_INVISIBLE 2
#define CLOSE_VISIBLE 3

/* character set list type */
#define G94         0 /* 94-character G-set */
#define G96         1 /* 96-character G-set */
#define MBG94       2 /* 94-character multibyte G-set */
#define MBG96       3 /* 96-character multibyte G-set */
#define COMPLETE    4 /* complete code */

/* character coding announcer */

#define BASIC7 0       /* basic 7-bit    */
#define BASIC8 1       /* basic 8-bit    */
#define EXTENDED7 2    /* extended 7-bit */
#define EXTENDED8 3    /* extended 8-bit */
